/*
 *  Copyright (C) 2002-2005 The SSI Project. All rights reserved.
 *  http://ssi.is.s.u-tokyo.ac.jp
 *
 *  This software is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY.  See the COPYING file for the precise
 *  terms and conditions for the use of this software.
 */
#include "libfftss.h"

#ifdef HAVE_X86_CPUID

int fftss_cpu = -1;

#ifdef HAVE_INTRIN_H
#define fftss_x86_cpuid(o, e) __cpuid(e, o)

#elif defined(_WIN64) && !defined(__INTEL_COMPILER)
/*
 *  Microsoft Visual Studio .NET 2003 does not support __asm keyword for EM64T/AMD64. 
 */
void fftss_x86_cpuid(int op, int *e_x) {
	fftss_cpu = FFTSS_X86_AMD|FFTSS_X86_MMX|FFTSS_X86_SSE|FFTSS_X86_SSE2;
};

#elif defined(_WIN32)
void fftss_x86_cpuid(int op, int *e_x)
{
    int ex[4];

	__asm {
		mov eax,op
		cpuid	
		mov dword ptr [ex+0],eax
		mov dword ptr [ex+4],ebx
		mov dword ptr [ex+8],ecx
		mov dword ptr [ex+12],edx
	}
	e_x[0] = ex[0];
	e_x[1] = ex[1];
	e_x[2] = ex[2];
	e_x[3] = ex[3];
}
#else
void fftss_x86_cpuid(int op, int *e_x)
{
  __asm__ ("cpuid"
	   : "=a" (e_x[0]),
	   "=b" (e_x[1]),
	   "=c" (e_x[2]),
	   "=d" (e_x[3])
	   : "a" (op));
}
#endif

void fftss_check_cpu(void)
{
  int e_x[4];

  fftss_cpu = 0;

  fftss_x86_cpuid(0, e_x);
/*  printf("%c\n", *(char *)e_x); */
  if (*(char *)e_x == 'G') fftss_cpu |= FFTSS_X86_INTEL;
  else fftss_cpu |= FFTSS_X86_AMD;

  fftss_x86_cpuid(1, e_x);
  if ((e_x[3] >> 23) & 1) fftss_cpu |= FFTSS_X86_MMX;
  if ((e_x[3] >> 25) & 1) fftss_cpu |= FFTSS_X86_SSE;
  if ((e_x[3] >> 26) & 1) fftss_cpu |= FFTSS_X86_SSE2;
  if ((e_x[2] >> 0) & 1) fftss_cpu |= FFTSS_X86_SSE3;
}

#else

int fftss_cpu = 0;
void fftss_check_cpu(void){};

#endif

